/*
 * MP3/MPlayer plugin to VDR (C++)
 *
 * (C) 2001-2003 Stefan Huelswitt <huels@iname.com>
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>

#include <mad.h>
#include <id3tag.h>

#include <vdr/player.h>
#include <vdr/device.h>
#include <vdr/thread.h>
#include <vdr/ringbuffer.h>
#include <vdr/tools.h>
#include <vdr/recording.h>

#include "common.h"
#include "setup-mp3.h"
#include "player-mp3.h"
#include "data-mp3.h"
#include "decoder.h"
#include "decoder-core.h"

// ----------------------------------------------------------------

#define OUT_BITS 16                                          // output 16 bit samples to DVB driver
#define OUT_FACT (OUT_BITS/8*2)                              // output factor is 16 bit & 2 channels -> 4 bytes
// cResample
#define MAX_NSAMPLES (1152*7)                                // max. buffer for resampled frame
// cNormalize
#define MAX_GAIN   3.0                                       // max. allowed gain
#define LIM_ACC    12                                        // bit, accuracy for lookup table
#define F_LIM_MAX  (mad_fixed_t)((1<<(MAD_F_FRACBITS+2))-1)  // max. value covered by lookup table
#define LIM_SHIFT  (MAD_F_FRACBITS-LIM_ACC)                  // shift value for table lookup
#define F_LIM_JMP  (mad_fixed_t)(1<<LIM_SHIFT)               // lookup table jump between values
// cLevel
#define POW_WIN 100                                          // window width for smoothing power values
#define EPSILON 0.00000000001                                // anything less than EPSILON is considered zero
// cMP3Player
#define MAX_FRAMESIZE   2048                                 // max. frame size allowed for DVB driver
#define HDR_SIZE        9
#define LPCM_SIZE       7
#define LEN_CORR        3
#define SPEEDCHECKSTART ((MP3BUFSIZE*1000/32000/2/2)+1000)   // play time to fill buffers before speed check starts (ms)
#define SPEEDCHECKTIME  3000                                 // play time for speed check (ms)

/*
struct LPCMHeader { int id:8;              // id
                    int frame_count:8;     // number of frames
                    int access_ptr:16;     // first acces unit pointer, i.e. start of audio frame
                    bool emphasis:1;       // audio emphasis on-off
                    bool mute:1;           // audio mute on-off
                    bool reserved:1;       // reserved
                    int frame_number:5;    // audio frame number
                    int quant_wlen:2;      // quantization word length
                    int sample_freq:2;     // audio sampling frequency (48khz=0, 96khz=1, 44,1khz=2, 32khz=3)
                    bool reserved2:1;      // reserved
                    int chan_count:3;      // number of audio channels - 1 (e.g. stereo = 1)
                    int dyn_range_ctrl:8;  // dynamic range control (0x80 if off)
                    };
*/

struct LPCMFrame {
  unsigned char PES[HDR_SIZE];
  unsigned char LPCM[LPCM_SIZE];
  unsigned char Data[MAX_FRAMESIZE-HDR_SIZE-LPCM_SIZE];
  };

// --- cResample ------------------------------------------------------------

// The resample code has been adapted from the madplay project
// (resample.c) found in the libmad distribution
   
class cResample {
private:
  mad_fixed_t ratio;
  mad_fixed_t step;
  mad_fixed_t last;
  mad_fixed_t resampled[MAX_NSAMPLES];
public:
  bool SetInputRate(unsigned int oldrate, unsigned int newrate);
  unsigned int ResampleBlock(unsigned int nsamples, const mad_fixed_t *old);
  const mad_fixed_t *Resampled(void) { return resampled; }
  };

bool cResample::SetInputRate(unsigned int oldrate, unsigned int newrate)
{
  if(oldrate<8000 || oldrate>newrate*6) { // out of range
    esyslog("WARNING: samplerate %d out of range 8000-%d\n",oldrate,newrate*6);
    return 0;
    }
  ratio=mad_f_tofixed((double)oldrate/(double)newrate);
  step=0; last=0;
#ifdef DEBUG
  static mad_fixed_t oldratio=0;
  if(oldratio!=ratio) {
    printf("mad: new resample ratio %f (from %d kHz to %d kHz)\n",mad_f_todouble(ratio),oldrate,newrate);
    oldratio=ratio;
    }
#endif
  return ratio!=MAD_F_ONE;
}

unsigned int cResample::ResampleBlock(unsigned int nsamples, const mad_fixed_t *old)
{
  // This resampling algorithm is based on a linear interpolation, which is
  // not at all the best sounding but is relatively fast and efficient.
  //
  // A better algorithm would be one that implements a bandlimited
  // interpolation.

  mad_fixed_t *nsam=resampled;
  const mad_fixed_t *end=old+nsamples;
  const mad_fixed_t *begin=nsam;

  if(step < 0) {
    step = mad_f_fracpart(-step);

    while (step < MAD_F_ONE) {
      *nsam++ = step ? last+mad_f_mul(*old-last,step) : last;
      step += ratio;
      if(((step + 0x00000080L) & 0x0fffff00L) == 0)
	step = (step + 0x00000080L) & ~0x0fffffffL;
      }
    step -= MAD_F_ONE;
    }

  while (end - old > 1 + mad_f_intpart(step)) {
    old += mad_f_intpart(step);
    step = mad_f_fracpart(step);
    *nsam++ = step ? *old + mad_f_mul(old[1] - old[0], step) : *old;
    step += ratio;
    if (((step + 0x00000080L) & 0x0fffff00L) == 0)
      step = (step + 0x00000080L) & ~0x0fffffffL;
    }

  if (end - old == 1 + mad_f_intpart(step)) {
    last = end[-1];
    step = -step;
    }
  else step -= mad_f_fromint(end - old);

  return nsam-begin;
}

// --- cLevel ----------------------------------------------------------------

// The normalize algorithm and parts of the code has been adapted from the
// Normalize 0.7 project. (C) 1999-2002, Chris Vaill <cvaill@cs.columbia.edu>

// A little background on how normalize computes the volume
// of a wav file, in case you want to know just how your
// files are being munged:
//
// The volumes calculated are RMS amplitudes, which corre
// spond (roughly) to perceived volume. Taking the RMS ampli
// tude of an entire file would not give us quite the measure
// we want, though, because a quiet song punctuated by short
// loud parts would average out to a quiet song, and the
// adjustment we would compute would make the loud parts
// excessively loud.
//
// What we want is to consider the maximum volume of the
// file, and normalize according to that. We break up the
// signal into 100 chunks per second, and get the signal
// power of each chunk, in order to get an estimation of
// "instantaneous power" over time. This "instantaneous
// power" signal varies too much to get a good measure of the
// original signal's maximum sustained power, so we run a
// smoothing algorithm over the power signal (specifically, a
// mean filter with a window width of 100 elements). The max
// imum point of the smoothed power signal turns out to be a
// good measure of the maximum sustained power of the file.
// We can then take the square root of the power to get maxi
// mum sustained RMS amplitude.

class cLevel {
private:
  double maxpow;
  mad_fixed_t peak;
  struct Power {
    // smooth
    int npow, wpow;
    double powsum, pows[POW_WIN];
    // sum
    unsigned int nsum;
    double sum;
    } power[2];
  //
  inline void AddPower(struct Power *p, double pow);
public:
  void Init(void);
  void GetPower(struct mad_pcm *pcm);
  double GetLevel(void);
  double GetPeak(void);
  };

void cLevel::Init(void)
{
  for(int l=0 ; l<2 ; l++) {
    struct Power *p=&power[l];
    p->sum=p->powsum=0.0; p->wpow=p->npow=p->nsum=0;
    for(int i=POW_WIN-1 ; i>=0 ; i--) p->pows[i]=0.0;
    }
  maxpow=0.0; peak=0;
}

void cLevel::GetPower(struct mad_pcm *pcm)
{
  for(int i=0 ; i<pcm->channels ; i++) {
    struct Power *p=&power[i];
    mad_fixed_t *data=pcm->samples[i];
    for(int n=pcm->length ; n>0 ; n--) {
      if(*data < -peak) peak = -*data;
      if(*data >  peak) peak =  *data;
      double s=mad_f_todouble(*data++);
      p->sum+=(s*s);
      if(++(p->nsum)>=pcm->samplerate/100) {
        AddPower(p,p->sum/(double)p->nsum);
        p->sum=0.0; p->nsum=0;
        }
      }
    }
}

void cLevel::AddPower(struct Power *p, double pow)
{
  p->powsum+=pow;
  if(p->npow>=POW_WIN) {
    if(p->powsum>maxpow) maxpow=p->powsum;
    p->powsum-=p->pows[p->wpow];
    }
  else p->npow++;
  p->pows[p->wpow]=pow;
  p->wpow=(p->wpow+1) % POW_WIN;
}

double cLevel::GetLevel(void)
{
  if(maxpow<EPSILON) {
    // Either this whole file has zero power, or was too short to ever
    // fill the smoothing buffer.  In the latter case, we need to just
    // get maxpow from whatever data we did collect.

    if(power[0].powsum>maxpow) maxpow=power[0].powsum;
    if(power[1].powsum>maxpow) maxpow=power[1].powsum;
    }
  double level=sqrt(maxpow/(double)POW_WIN);     // adjust for the smoothing window size and root
  d(printf("norm: new volumen level=%f peak=%f\n",level,mad_f_todouble(peak)))
  return level;
}

double cLevel::GetPeak(void)
{
  return mad_f_todouble(peak);
}

// --- cNormalize ------------------------------------------------------------

class cNormalize {
private:
  mad_fixed_t gain;
  double d_limlvl, one_limlvl;
  mad_fixed_t limlvl;
  bool dogain, dolimit;
#ifdef DEBUG
  // stats
  unsigned long limited, clipped, total;
  mad_fixed_t peak;
#endif
  // limiter
#ifdef USE_FAST_LIMITER
  mad_fixed_t *table, tablestart;
  int tablesize;
  inline mad_fixed_t FastLimiter(mad_fixed_t x);
#endif
  inline mad_fixed_t Limiter(mad_fixed_t x);
public:
  cNormalize(void);
  ~cNormalize();
  void Init(double Level, double Peak);
  void Stats(void);
  void AddGain(struct mad_pcm *pcm);
  };

cNormalize::cNormalize(void)
{
  d_limlvl=(double)MP3Setup.LimiterLevel/100.0;
  one_limlvl=1-d_limlvl;
  limlvl=mad_f_tofixed(d_limlvl);
  d(printf("norm: lim_lev=%f lim_acc=%d\n",d_limlvl,LIM_ACC))

#ifdef USE_FAST_LIMITER
  mad_fixed_t start=limlvl & ~(F_LIM_JMP-1);
  tablestart=start;
  tablesize=(unsigned int)(F_LIM_MAX-start)/F_LIM_JMP + 2;
  table=new mad_fixed_t[tablesize];
  if(table) {
    d(printf("norm: table size=%d start=%08x jump=%08x\n",tablesize,start,F_LIM_JMP))
    for(int i=0 ; i<tablesize ; i++) {
      table[i]=Limiter(start);
      start+=F_LIM_JMP;
      }
    tablesize--; // avoid a -1 in FastLimiter()

    // do a quick accuracy check, just to be sure that FastLimiter() is working
    // as expected :-)
#ifdef ACC_DUMP
    FILE *out=fopen("/tmp/limiter","w");
#endif
    mad_fixed_t maxdiff=0;
    for(mad_fixed_t x=F_LIM_MAX ; x>=limlvl ; x-=mad_f_tofixed(1e-4)) {
      mad_fixed_t diff=mad_f_abs(Limiter(x)-FastLimiter(x));
      if(diff>maxdiff) maxdiff=diff;
#ifdef ACC_DUMP
      fprintf(out,"%0.10f\t%0.10f\t%0.10f\t%0.10f\t%0.10f\n",
        mad_f_todouble(x),mad_f_todouble(Limiter(x)),mad_f_todouble(FastLimiter(x)),mad_f_todouble(diff),mad_f_todouble(maxdiff));
      if(ferror(out)) break;
#endif
      }
#ifdef ACC_DUMP
    fclose(out);
#endif
    d(printf("norm: accuracy %.12f\n",mad_f_todouble(maxdiff)))
    if(mad_f_todouble(maxdiff)>1e-6) {
      esyslog("ERROR: accuracy check failed, normalizer disabled");
      delete table; table=0;
      }
    }
  else esyslog("ERROR: no memory for lookup table, normalizer disabled");
#endif // USE_FAST_LIMITER
}

cNormalize::~cNormalize()
{
#ifdef USE_FAST_LIMITER
  delete table;
#endif
}

void cNormalize::Init(double Level, double Peak)
{
  double Target=(double)MP3Setup.TargetLevel/100.0;
  double dgain=Target/Level;
  if(dgain>MAX_GAIN) dgain=MAX_GAIN;
  gain=mad_f_tofixed(dgain);
  // Check if we actually need to apply a gain
  dogain=(Target>0.0 && fabs(1-dgain)>MIN_GAIN);
#ifdef USE_FAST_LIMITER
  if(!table) dogain=false;
#endif
  // Check if we actually need to do limiting:
  // we have to if limiter is enabled, if gain>1 and if the peaks will clip.
  dolimit=(d_limlvl<1.0 && dgain>1.0 && Peak*dgain>1.0);
#ifdef DEBUG
  printf("norm: gain=%f dogain=%d dolimit=%d (target=%f level=%f peak=%f)\n",dgain,dogain,dolimit,Target,Level,Peak);
  limited=clipped=total=0; peak=0;
#endif
}

void cNormalize::Stats(void)
{
#ifdef DEBUG
  if(total)
    printf("norm: stats tot=%ld lim=%ld/%.3f%% clip=%ld/%.3f%% peak=%.3f\n",
           total,limited,(double)limited/total*100.0,clipped,(double)clipped/total*100.0,mad_f_todouble(peak));
#endif
}

mad_fixed_t cNormalize::Limiter(mad_fixed_t x)
{
// Limiter function:
//
//        / x                                                (for x <= lev)
//   x' = |
//        \ tanh((x - lev) / (1-lev)) * (1-lev) + lev        (for x > lev)
//
// call only with x>=0. For negative samples, preserve sign outside this function
//
// With limiter level = 0, this is equivalent to a tanh() function;
// with limiter level = 1, this is equivalent to clipping.

  if(x>limlvl) {
#ifdef DEBUG
    if(x>MAD_F_ONE) clipped++;
    limited++;
#endif
    x=mad_f_tofixed(tanh((mad_f_todouble(x)-d_limlvl) / one_limlvl) * one_limlvl + d_limlvl);
    }
  return x;
}

#ifdef USE_FAST_LIMITER
mad_fixed_t cNormalize::FastLimiter(mad_fixed_t x)
{
// The fast algorithm is based on a linear interpolation between the
// the values in the lookup table. Relays heavly on libmads fixed point format.

  if(x>limlvl) {
    int i=(unsigned int)(x-tablestart)/F_LIM_JMP;
#ifdef DEBUG
    if(x>MAD_F_ONE) clipped++;
    limited++;
    if(i>=tablesize) printf("norm: overflow x=%f x-ts=%f i=%d tsize=%d\n",
                            mad_f_todouble(x),mad_f_todouble(x-tablestart),i,tablesize);
#endif
    mad_fixed_t r=x & (F_LIM_JMP-1);
    x=MAD_F_ONE;
    if(i<tablesize) {
      mad_fixed_t *ptr=&table[i];
      x=*ptr;
      mad_fixed_t d=*(ptr+1)-x;
      //x+=mad_f_mul(d,r)<<LIM_ACC;                // this is not accurate as mad_f_mul() does >>MAD_F_FRACBITS
                                                   // which is senseless in the case of following <<LIM_ACC.
      x+=((long long)d*(long long)r)>>LIM_SHIFT;   // better, don't know if works on all machines
      }
    }
  return x;
}
#endif

#ifdef USE_FAST_LIMITER
#define LIMITER_FUNC FastLimiter
#else
#define LIMITER_FUNC Limiter
#endif

void cNormalize::AddGain(struct mad_pcm *pcm)
{
  if(dogain) {
    for(int i=0 ; i<pcm->channels ; i++) {
      mad_fixed_t *data=pcm->samples[i];
#ifdef DEBUG
      total+=pcm->length;
#endif
      if(dolimit) {
        for(int n=pcm->length ; n>0 ; n--) {
          mad_fixed_t s=mad_f_mul(*data,gain);
          if(s<0) {
            s=-s;
#ifdef DEBUG
            if(s>peak) peak=s;
#endif
            s=LIMITER_FUNC(s);
            s=-s;
            }
          else {
#ifdef DEBUG
            if(s>peak) peak=s;
#endif
            s=LIMITER_FUNC(s);
            }
          *data++=s;
          }
        }
      else {
        for(int n=pcm->length ; n>0 ; n--) {
          mad_fixed_t s=mad_f_mul(*data,gain);
#ifdef DEBUG
          if(s>peak) peak=s;
          else if(-s>peak) peak=-s;
#endif
          if(s>MAD_F_ONE) s=MAD_F_ONE;   // do clipping
          if(s<-MAD_F_ONE) s=-MAD_F_ONE;
          *data++=s;
          }
        }
      }
    }
}

// --- cScale ----------------------------------------------------------------

// The dither code has been adapted from the madplay project
// (audio.c) found in the libmad distribution

enum eAudioMode { amRound, amDither };

class cScale {
private:
  enum { MIN=-MAD_F_ONE, MAX=MAD_F_ONE - 1 };
#ifdef DEBUG
  // audio stats
  unsigned long clipped_samples;
  mad_fixed_t peak_clipping;
  mad_fixed_t peak_sample;
#endif
  // dither
  struct dither {
    mad_fixed_t error[3];
    mad_fixed_t random;
    } leftD, rightD;
  //
  inline mad_fixed_t Clip(mad_fixed_t sample, bool stats=true);
  inline signed long LinearRound(mad_fixed_t sample);
  inline unsigned long Prng(unsigned long state);
  inline signed long LinearDither(mad_fixed_t sample, struct dither *dither);
public:
  void Init(void);
  void Stats(void);
  unsigned int ScaleBlock(unsigned char *data, unsigned int size, unsigned int &nsamples, const mad_fixed_t * &left, const mad_fixed_t * &right, eAudioMode mode);
  };

void cScale::Init(void)
{
#ifdef DEBUG
  clipped_samples=0; peak_clipping=peak_sample=0;
#endif
  memset(&leftD,0,sizeof(leftD));
  memset(&rightD,0,sizeof(rightD));
}

void cScale::Stats(void)
{
#ifdef DEBUG
  printf("mp3: scale stats clipped=%ld peak_clip=%f peak=%f\n",
         clipped_samples,mad_f_todouble(peak_clipping),mad_f_todouble(peak_sample));
#endif
}

// gather signal statistics while clipping
mad_fixed_t cScale::Clip(mad_fixed_t sample, bool stats)
{
#ifndef DEBUG
  if (sample > MAX) sample = MAX;
  if (sample < MIN) sample = MIN;
#else
  if(!stats) {
    if (sample > MAX) sample = MAX;
    if (sample < MIN) sample = MIN;
    }
  else {
    if (sample >= peak_sample) {
      if (sample > MAX) {
        ++clipped_samples;
        if (sample - MAX > peak_clipping)
	  peak_clipping = sample - MAX;
        sample = MAX;
        }
      peak_sample = sample;
      }
    else if (sample < -peak_sample) {
      if (sample < MIN) {
        ++clipped_samples;
        if (MIN - sample > peak_clipping)
	  peak_clipping = MIN - sample;
        sample = MIN;
        }
      peak_sample = -sample;
      }
    }
#endif
  return sample;
}

// generic linear sample quantize routine
signed long cScale::LinearRound(mad_fixed_t sample)
{
  // round
  sample += (1L << (MAD_F_FRACBITS - OUT_BITS));
  // clip
  sample=Clip(sample);
  // quantize and scale
  return sample >> (MAD_F_FRACBITS + 1 - OUT_BITS);
}

// 32-bit pseudo-random number generator
unsigned long cScale::Prng(unsigned long state)
{
  return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

// generic linear sample quantize and dither routine
signed long cScale::LinearDither(mad_fixed_t sample, struct dither *dither)
{
  unsigned int scalebits;
  mad_fixed_t output, mask, random;

  // noise shape
  sample += dither->error[0] - dither->error[1] + dither->error[2];
  dither->error[2] = dither->error[1];
  dither->error[1] = dither->error[0] / 2;
  // bias
  output = sample + (1L << (MAD_F_FRACBITS + 1 - OUT_BITS - 1));
  scalebits = MAD_F_FRACBITS + 1 - OUT_BITS;
  mask = (1L << scalebits) - 1;
  // dither
  random  = Prng(dither->random);
  output += (random & mask) - (dither->random & mask);
  dither->random = random;
  // clip
  output=Clip(output);
  sample=Clip(sample,false);
  // quantize
  output &= ~mask;
  // error feedback
  dither->error[0] = sample - output;
  // scale
  return output >> scalebits;
}

// write a block of signed 16-bit big-endian PCM samples
unsigned int cScale::ScaleBlock(unsigned char *data, unsigned int size, unsigned int &nsamples, const mad_fixed_t * &left, const mad_fixed_t * &right, eAudioMode mode)
{
  signed int sample;
  unsigned int len, res;

  len=size/OUT_FACT; res=size;
  if(len>nsamples) { len=nsamples; res=len*OUT_FACT; }
  nsamples-=len;

  if(right) {  // stereo
    switch (mode) {
      case amRound:
        while (len--) {
          sample  = LinearRound(*left++);
          *data++ = sample >> 8;
          *data++ = sample >> 0;
          sample  = LinearRound(*right++);
          *data++ = sample >> 8;
          *data++ = sample >> 0;
          }
        break;
      case amDither:
        while (len--) {
	  sample  = LinearDither(*left++,&leftD);
	  *data++ = sample >> 8;
	  *data++ = sample >> 0;
	  sample  = LinearDither(*right++,&rightD);
	  *data++ = sample >> 8;
	  *data++ = sample >> 0;
          }
        break;
      }
    }
  else {  // mono, duplicate left channel
    switch (mode) {
      case amRound:
        while (len--) {
	  sample  = LinearRound(*left++);
	  *data++ = sample >> 8;
	  *data++ = sample >> 0;
	  *data++ = sample >> 8;
	  *data++ = sample >> 0;
          }
        break;
      case amDither:
        while (len--) {
	  sample  = LinearDither(*left++,&leftD);
	  *data++ = sample >> 8;
	  *data++ = sample >> 0;
	  *data++ = sample >> 8;
	  *data++ = sample >> 0;
          }
        break;
      }
    }
 return res;
}

// --- cBackgroundScan ---------------------------------------------------------

class cBackgroundScan : public cThread {
private:
  cPlayList *pl;
  bool stop, throttle;
  //
  virtual void Action(void);
public:
  cBackgroundScan(cPlayList *list);
  ~cBackgroundScan();
  inline void Throttle(void);
  };

cBackgroundScan::cBackgroundScan(cPlayList *list)
{
  pl=list; stop=throttle=false;
  Start();
}

cBackgroundScan::~cBackgroundScan()
{
  stop=true;
  Cancel(4);
}

void cBackgroundScan::Throttle(void)
{
  throttle=true;
}

void cBackgroundScan::Action(void)
{
  d(printf("mp3: background scan thread started (pid=%d)\n", getpid()))
  nice(5);
  cSong *song=pl->First();
  while(!stop && song) {
    if(throttle) { throttle=false; sleep(1); continue; }
    song->Info(true);
    song=pl->cList<cSong>::Next(song);
    }
  d(printf("mp3: background scan thread ended (pid=%d)\n", getpid()))
}

// --- cMP3Player ----------------------------------------------------------

class cMP3Player : public cPlayer, cThread {
private:
  bool active, started;
  cRingBufferFrame *ringBuffer;
  cMutex playModeMutex;
  cCondVar playModeCond;
//
  cPlayList *playlist;
  cSong *current, *playing;
  cBackgroundScan *scan;
//
  cMutex infoMutex;
  cSongInfo *currentSi;
//
  cDecoder *decoder;
  cFrame *rframe, *pframe;
  enum { tmNone=0, tmLoop=1, tmShuffle=2 };
  int trickmode;
  enum ePlayMode { pmPlay, pmStopped, pmPaused, pmStartup };
  ePlayMode playMode;
  enum eState { msStart, msStop, msDecode, msNormalize, msResample, msOutput, msError, msEof, msWait };
  eState state;
  bool levelgood, isStream;
  unsigned int dvbSampleRate;
//
  int *shuffle;
  int index, num, maxnum;
//
  void Empty(void);
  bool NextFile(void);
  bool PrevFile(void);
  void InitShuffle(void);
  void StopPlay(void);
  void InitPlaylist(void);
  void SetPlayMode(ePlayMode mode);
  void WaitPlayMode(ePlayMode mode, bool inv);
protected:
  virtual void Activate(bool On);
  virtual void Action(void);
public:
  cMP3Player(cPlayList *plist);
  virtual ~cMP3Player();
//
  bool Active(void) { return active; }
  void Pause(void);
  void Play(void);
  void Forward(void);
  void Backward(void);
  void Goto(int Index, bool Still=false);
  void SkipSeconds(int secs);
  void ToggleShuffle(void);
  void ToggleLoop(void);
  virtual bool GetIndex(int &Current, int &Total, bool SnapToIFrame=false);
  bool GetPlayInfo(cMP3PlayInfo *rm);
  void NewPlaylist(cPlayList *plist);
  bool IsStream(void) { return isStream; }
  };

cMP3Player::cMP3Player(cPlayList *plist)
:cPlayer(MP3Setup.BackgrMode?pmAudioOnly:pmAudioOnlyBlack)
{
  playlist=plist;

  active=true; started=false; isStream=false;
  ringBuffer=new cRingBufferFrame(MP3BUFSIZE);
  rframe=0; pframe=0; decoder=0;
  playMode=pmStartup; state=msStop;

  index=0; playing=0; currentSi=0;
  shuffle=0; scan=0;
  
  trickmode=tmNone;
  if(MP3Setup.InitLoopMode) trickmode|=tmLoop;
  if(MP3Setup.InitShuffleMode) trickmode|=tmShuffle;
  d(printf("mp3: inital trickmode=%d\n",trickmode))
}

cMP3Player::~cMP3Player()
{
  Detach();
  delete scan;
  delete playlist;
  delete shuffle;
  delete ringBuffer;
}

void cMP3Player::Activate(bool On)
{
  if(On) {
    if(playlist && !started) {
      playMode=pmStartup; Start(); started=true;
      InitPlaylist();
      playModeMutex.Lock();
      WaitPlayMode(pmStartup,true); // wait for the decoder to become ready
      playModeMutex.Unlock();
      Lock();
      if(NextFile()) Play();
      Unlock();
      }
    }
  else if(started && active) {
      Lock(); StopPlay(); Unlock();
      active=false;
      SetPlayMode(pmStartup);
      Cancel(2);
    }
}

void cMP3Player::NewPlaylist(cPlayList *plist)
{
  d(printf("mp3: got request for new playlist\n"))
  Lock(); StopPlay(); Unlock();
  delete scan; scan=0;
  delete playlist;
  playlist=plist;
  InitPlaylist();
  if(NextFile()) Play();
}

void cMP3Player::InitPlaylist(void)
{
  maxnum=playlist->Count();
  d(printf("mp3: playlist has %d files, first is %s\n",maxnum,playlist->First()->Name()))
  delete shuffle;
  shuffle=new int[maxnum];
  InitShuffle();
  current=0;
  if(MP3Setup.Id3PreScan) {
    if(scan) delete scan;
    scan=new cBackgroundScan(playlist);
    }
}

void cMP3Player::InitShuffle(void)
{
  int i, ran, t;

  srand(time(0));
  for(i=0 ; i<maxnum ; i++) shuffle[i]=i;
  if(maxnum >= 2)
    for(i=0 ; i<maxnum ; i++) {
      ran=(rand() % (maxnum*4-4))/4; ran+=(ran >= i);
      t=shuffle[i]; shuffle[i]=shuffle[ran]; shuffle[ran]=t;
      }
#ifdef DEBUG
  printf("mp3: shuffle order ");
  for(i=0 ; i<maxnum ; i++) printf("%d ",shuffle[i]);
  printf("\n");
#endif
}

void cMP3Player::SetPlayMode(ePlayMode mode)
{
  playModeMutex.Lock();
  if(mode!=playMode) {
    playMode=mode;
    d(printf("mp3: setting mode=%d (pid=%d)\n",mode,getpid()))
    playModeCond.Broadcast();
    }
  playModeMutex.Unlock();
}

void cMP3Player::WaitPlayMode(ePlayMode mode, bool inv)
{
  // must be called with playModeMutex LOCKED !!!

  while(active && ((!inv && mode!=playMode) || (inv && mode==playMode))) {
    d(printf("mp3: entering wait for mode%s%d with mode=%d (pid=%d)\n",inv?"!=":"==",mode,playMode,getpid()))
    playModeCond.Wait(playModeMutex);
    d(printf("mp3: returning from wait with mode=%d (pid=%d)\n",playMode,getpid()))
    }
}

void cMP3Player::Action(void)
{
  struct Decode *ds=0;
  struct mad_pcm *pcm=0;
  cResample resample[2];
  unsigned int nsamples[2];
  const mad_fixed_t *data[2];
  cScale scale;
  cLevel level;
  cNormalize norm;
  bool haslevel=false;
  struct LPCMFrame lpcmFrame;
  const unsigned char *p=0;
  int pc=0, only48khz=MP3Setup.Only48kHz;
  cPoller poll;
#ifdef DEBUG
  int beat=0;
#endif

  dsyslog("mp3: player thread started (pid=%d)", getpid());

  memset(&lpcmFrame,0,sizeof(lpcmFrame));
  lpcmFrame.PES[2]=0x01;
  lpcmFrame.PES[3]=0xbd;
  lpcmFrame.PES[6]=0x87;
  lpcmFrame.LPCM[0]=0xa0; // substream ID
  lpcmFrame.LPCM[1]=0xff;
  lpcmFrame.LPCM[5]=0x01;
  lpcmFrame.LPCM[6]=0x80;

  dvbSampleRate=48000;
  state=msStop;
  SetPlayMode(pmStopped);
 
  while(active) {
#ifdef DEBUG
    if(time(0)>=beat+30) {
      printf("mp3: heartbeat buffer=%d\n",ringBuffer->Available());
      scale.Stats(); if(haslevel) norm.Stats();
      beat=time(0);
      }
#endif

    Lock();

    if(!rframe && playMode==pmPlay) {
      switch(state) {
        case msStart:
          d(printf("mp3: starting play\n"))
          index=0; playing=current;
          if((decoder=playing->Decoder()) && decoder->Start()) {
            isStream=decoder->IsStream(); levelgood=!isStream; haslevel=false;
            cSongInfo *si=playing->Info(false);
            if(si && si->Level>0.0) {
              d(printf("mp3: found song level=%f peak=%f\n",si->Level,si->Peak))
              haslevel=true;
              norm.Init(si->Level,si->Peak);
              }
            d(printf("mp3: isStream=%d levelgood=%d haslevel=%d\n",isStream,levelgood,haslevel))
            infoMutex.Lock(); currentSi=si; infoMutex.Unlock();
            scale.Init();
            level.Init();
            state=msDecode;
            break;
            }
          esyslog("ERROR: playlist entry %s is not a valid file",current->Name());
          state=msEof;
          break;
        case msDecode:
          ds=decoder->Decode();
          switch(ds->status) {
            case dsPlay:
              pcm=ds->pcm;
              index=ds->index/1000;
              state=msNormalize;
              break;
            case dsSkip:
            case dsSoftError:
              // skipping, state unchanged, next decode
              break;
            case dsEof:
              if(!haslevel && levelgood) {   // save level & peak to infocache on eof
                double l=level.GetLevel();
                if(l>0.0) {
                  cSongInfo *si=decoder->SongInfo(false);
                  cFileInfo *fi=decoder->FileInfo();
                  if(si && fi) {
                    si->Level=l;
                    si->Peak=level.GetPeak();
                    InfoCache.Cache(si,fi,true);
                    }
                  }
                }
              state=msEof;
              break;
            case dsOK:
            case dsError:
              state=msError;
              break;
            }
          break;
        case msNormalize:
          if(!haslevel) { if(levelgood) level.GetPower(pcm); }
          else norm.AddGain(pcm);
          state=msResample;
          break;
        case msResample:
#ifdef DEBUG
          {
          static unsigned int oldrate=0;
          if(oldrate!=pcm->samplerate) {
            printf("mp3: new input sample rate %d\n",pcm->samplerate);
            oldrate=pcm->samplerate;
            }
          }
#endif
          nsamples[0]=nsamples[1]=pcm->length;
          data[0]=pcm->samples[0];
          data[1]=pcm->channels>1 ? pcm->samples[1]:0;

          lpcmFrame.LPCM[5]&=0xcf;
          dvbSampleRate=48000;
          if(!only48khz) {
            switch(pcm->samplerate) {    // If one of the supported frequencies, do it without resampling.
              case 96000:                // Select a "even" upsampling frequency if possible, too.
                lpcmFrame.LPCM[5]|=1<<4;
                dvbSampleRate=96000;
                break;
              //case 48000: // this is already the default ...
              //  lpcmFrame.LPCM[5]|=0<<4;
              //  dvbSampleRate=48000;
              //  break;
              case 11025:
              case 22050:
              case 44100:
                lpcmFrame.LPCM[5]|=2<<4;
                dvbSampleRate=44100;
                break;
              case 8000:
              case 16000:
              case 32000:
                lpcmFrame.LPCM[5]|=3<<4;
                dvbSampleRate=32000;
                break;
              }
            }

          if(dvbSampleRate!=pcm->samplerate) {
            if(resample[0].SetInputRate(pcm->samplerate,dvbSampleRate)) {
              nsamples[0]=resample[0].ResampleBlock(nsamples[0],data[0]);
              data[0]    =resample[0].Resampled();
              }
            if(data[1] && resample[1].SetInputRate(pcm->samplerate,dvbSampleRate)) {
              nsamples[1]=resample[1].ResampleBlock(nsamples[1],data[1]);
              data[1]    =resample[1].Resampled();
              }
            }
          state=msOutput;
          break;
        case msOutput:
          if(nsamples[0]>0) {
            unsigned int outlen=scale.ScaleBlock(lpcmFrame.Data,sizeof(lpcmFrame.Data),nsamples[0],data[0],data[1],MP3Setup.AudioMode?amDither:amRound);
            if(outlen) { 
              outlen += sizeof(lpcmFrame.LPCM)+LEN_CORR;
              lpcmFrame.PES[4]=outlen>>8;
              lpcmFrame.PES[5]=outlen;
              rframe = new cFrame((unsigned char *)&lpcmFrame,outlen+sizeof(lpcmFrame.PES)-LEN_CORR);
              }
            }
          else state=msDecode;
          break;
        case msError:
        case msEof:
          d(printf("mp3: eof or error\n"))
          if(NextFile()) {
            state=msStart;
            d(printf("mp3: playing next %s\n",current->Name()))
            }
          else {
            state=msWait;
            d(printf("mp3: end of playlist\n"))
            }
          // fall through
        case msStop:
          d(printf("mp3: stopping play\n"))
          infoMutex.Lock(); currentSi=0; infoMutex.Unlock();
          playing=0;
          if(decoder) { decoder->Stop(); decoder=0; }
          levelgood=false;
          scale.Stats(); if(haslevel) norm.Stats();
          if(state==msStop) SetPlayMode(pmStopped);
          break;
        case msWait:
          if(ringBuffer->Available()==0) {
            active=false;
            SetPlayMode(pmStopped);
            }
          break;
        }
      }

    if(rframe && ringBuffer->Put(rframe)) rframe=0;

    if(!pframe && playMode==pmPlay) {
      pframe=ringBuffer->Get();
      if(pframe) {
        p=pframe->Data();
        pc=pframe->Count();
        }
      }

    if(pframe) {
      int w=PlayVideo(p,pc);
      if(w>0) {
        p+=w; pc-=w;
        if(pc<=0) {
          ringBuffer->Drop(pframe);
          pframe=0;
          }
        }
      else if(w<0 && FATALERRNO) {
        LOG_ERROR;
        break;
        }
      }

    Unlock();

    if(ringBuffer->Available()<(MP3BUFSIZE*70/100) && scan)
      scan->Throttle(); // throttle background scanner if buffer is low

    if((rframe || state==msWait) && pframe) {
      // Wait for output to become ready
      DevicePoll(poll,500);
      }
    else if(playMode!=pmPlay) {
      playModeMutex.Lock();
      if(playMode!=pmPlay) WaitPlayMode(playMode,true);  // Wait on playMode change
      playModeMutex.Unlock();
      }
    }

  Lock();
  delete rframe; rframe=0;
  if(decoder) { decoder->Stop(); decoder=0; }
  playing=0;
  SetPlayMode(pmStopped);
  Unlock();
  active=false;

  dsyslog("mp3: player thread ended (pid=%d)", getpid());
}

void cMP3Player::Empty(void)
{
  Lock();
  delete rframe; rframe=0; pframe=0;
  ringBuffer->Clear();
  DeviceClear();
  Unlock();
}

void cMP3Player::StopPlay(void) // StopPlay() must be called in locked state!!!
{
  if(playMode!=pmStopped) {
    Empty();
    state=msStop;
    SetPlayMode(pmPlay);
    Unlock();                 // let the decode thread process the stop signal
    playModeMutex.Lock();
    WaitPlayMode(pmStopped,false);
    playModeMutex.Unlock();
    Lock();
    }
}

bool cMP3Player::NextFile(void)
{
  cSong *newcurr;
  int nnum=0;
  bool res=false;

  if(trickmode & tmShuffle) {
    newcurr=0;
    if(current) nnum=num+1; else nnum=1;
    if(nnum>maxnum && (trickmode & tmLoop)) { nnum=1; InitShuffle(); }
    if(nnum<=maxnum) newcurr=playlist->Get(shuffle[nnum-1]);
    }
  else {
    if(current) {
      newcurr=playlist->cList<cSong>::Next(current);
      if(!newcurr && (trickmode & tmLoop)) newcurr=playlist->First();
      }
    else newcurr=playlist->First();
    }

  if(newcurr) {
    current=newcurr; num=nnum;
    res=true;
    }
  return res;
}

bool cMP3Player::PrevFile(void)
{
  cSong *newcurr=0;
  int nnum=0;
  bool res=false;

  if(index>=2) { newcurr=current; nnum=num; }
  else {
    if(trickmode & tmShuffle) {
      if(current) {
        nnum=num-1;
        if(nnum>=1) newcurr=playlist->Get(shuffle[nnum-1]);
        }
      }
    else {
      if(current) {
        newcurr=playlist->cList<cSong>::Prev(current);
        }
      }
    }

  if(newcurr) {
    current=newcurr; num=nnum;
    res=true;
    }
  return res;
}

void cMP3Player::Pause(void)
{
  if(playMode==pmPaused) Play();
  else if(playMode==pmPlay && !isStream) {
    d(printf("mp3: pause\n"))
//    DeviceFreeze();
    SetPlayMode(pmPaused);
    }
}

void cMP3Player::Play(void)
{
  Lock();
  if(playMode!=pmPlay && current) {
    d(printf("mp3: playing %s\n",current->Name()))
    if(playMode==pmStopped) state=msStart;
//    DevicePlay();
    SetPlayMode(pmPlay);
    }
  Unlock();
}

void cMP3Player::Forward(void)
{
  d(printf("mp3: skip forward\n"))
  Lock();
  if(NextFile()) { StopPlay(); Play(); }
  Unlock();
}

void cMP3Player::Backward(void)
{
  d(printf("mp3: skip backward\n"))
  Lock();
  if(PrevFile()) { StopPlay(); Play(); }
  Unlock();
}

void cMP3Player::Goto(int Index, bool Still)
{
  if(Index>0 && Index<=maxnum) {
    Lock();
    StopPlay();
    current=playlist->Get(Index-1);
    Play();
    Unlock();
    }
}

void cMP3Player::SkipSeconds(int secs)
{
  if(playMode!=pmStopped && !isStream) {
    Lock();
    if(playMode==pmPaused) SetPlayMode(pmPlay);
    if(decoder && decoder->Skip(secs,ringBuffer->Available(),dvbSampleRate))
      levelgood=false;
    Empty();
    Unlock();
    }
}

void cMP3Player::ToggleShuffle(void)
{
  trickmode ^= tmShuffle;
  d(printf("mp3: shuffle mode toggled : %d\n",trickmode & tmShuffle))
  if(trickmode & tmShuffle) {
    Lock();
    StopPlay();
    current=0;
    if(NextFile()) Play();
    Unlock();
    }
}

void cMP3Player::ToggleLoop(void)
{
  trickmode ^= tmLoop;
  d(printf("mp3: loop mode toggled : %d\n",trickmode & tmLoop))
}

bool cMP3Player::GetIndex(int &Current, int &Total, bool SnapToIFrame)
{
  bool res=false;
  Current=SecondsToFrames(index); Total=-1;
  infoMutex.Lock();
  if(currentSi && currentSi->HasInfo()) { Total=SecondsToFrames(currentSi->Total); res=true; }
  infoMutex.Unlock();
  return res;
}

bool cMP3Player::GetPlayInfo(cMP3PlayInfo *rm)
{
  static char *modestr[] = { "Mono","Dual","Joint-Stereo","Stereo" };
  bool res=false;

  rm->PlayList=playlist;
  rm->Num=current->Index()+1; rm->MaxNum=maxnum;
  rm->Loop=trickmode & tmLoop; rm->Shuffle=trickmode & tmShuffle;
  infoMutex.Lock();
  if(currentSi && currentSi->HasInfo()) {
    if(currentSi->Title) strn0cpy(rm->Title,currentSi->Title,sizeof(rm->Title));
    else strn0cpy(rm->Title,current->Name(),sizeof(rm->Title));
    rm->Artist[0]=0; if(currentSi->Artist) strn0cpy(rm->Artist,currentSi->Artist,sizeof(rm->Artist));
    rm->Album[0]=0; if(currentSi->Album) strn0cpy(rm->Album,currentSi->Album,sizeof(rm->Album));
    strn0cpy(rm->SMode,modestr[currentSi->ChMode],sizeof(rm->SMode));
    rm->Year=currentSi->Year;
    rm->SampleFreq=currentSi->SampleFreq; rm->Bitrate=currentSi->Bitrate; rm->MaxBitrate=currentSi->MaxBitrate;
    res=true;
    }
  infoMutex.Unlock();
  return res;
}
// --- cMP3PlayerControl -------------------------------------------------------

cMP3PlayerControl::cMP3PlayerControl(cPlayList *plist)
:cControl(mp3player = new cMP3Player(plist))
{}

cMP3PlayerControl::~cMP3PlayerControl()
{
  Stop();
}

bool cMP3PlayerControl::Active(void)
{
  return mp3player && mp3player->Active();
}

void cMP3PlayerControl::Stop(void)
{
  delete mp3player; mp3player=0;
}

void cMP3PlayerControl::Pause(void)
{
  if(mp3player) mp3player->Pause();
}

void cMP3PlayerControl::Play(void)
{
  if(mp3player) mp3player->Play();
}

void cMP3PlayerControl::Forward(void)
{
  if(mp3player) mp3player->Forward();
}

void cMP3PlayerControl::Backward(void)
{
  if(mp3player) mp3player->Backward();
}

void cMP3PlayerControl::SkipSeconds(int Seconds)
{
  if(mp3player) mp3player->SkipSeconds(Seconds);
}

void cMP3PlayerControl::Goto(int Position, bool Still)
{
  if(mp3player) mp3player->Goto(Position, Still);
}

void cMP3PlayerControl::ToggleShuffle(void)
{
  if(mp3player) mp3player->ToggleShuffle();
}

void cMP3PlayerControl::ToggleLoop(void)
{
  if(mp3player) mp3player->ToggleLoop();
}

bool cMP3PlayerControl::GetPlayInfo(cMP3PlayInfo *rm)
{
  return mp3player && mp3player->GetPlayInfo(rm);
}

void cMP3PlayerControl::NewPlaylist(cPlayList *plist)
{
  if(mp3player) mp3player->NewPlaylist(plist);
}

bool cMP3PlayerControl::IsStream(void)
{
  if(mp3player) return mp3player->IsStream();
  return true;
}
